from .base_segmenter import BaseSegmenter


class LongestStringMatching(BaseSegmenter):
    """Longest string matching segmenter.

    This model constructs predicted words by moving from left to right
    along an unsegmented sentence and finding the longest matching words,
    constrained by a maximum word length parameter.
    """

    def __init__(self, *, max_word_length):
        """Initialize a longest string matching segmenter.

        Parameters
        ----------
        max_word_length : int
            Maximum word length in the segmented sentences during prediction

        Raises
        ------
        ValueError
            If max_word_length is < 2.
        """
        self.max_word_length = max_word_length
        if self.max_word_length < 2:
            raise ValueError(
                f"max_word_length must be >= 2 to be meaningful: {max_word_length}"
            )

    def fit(self, sents):
        self._words = set()
        for sent in sents:
            # Don't waste memory for words of length 1, which are practically
            # useless for this algorithm.
            self._words |= set(word for word in sent if len(word) > 1)

    def _predict_sent(self, sent_str):
        sent_predicted = []
        max_word_length = min(len(sent_str), self.max_word_length)

        i = 0
        j = i + max_word_length

        while i < len(sent_str):
            j = min(j, len(sent_str))
            test = sent_str[i:j]

            if len(test) == 1 or test in self._words:
                sent_predicted.append(test)
                i = j
                j = i + max_word_length
            else:
                j -= 1

        return sent_predicted



import random

from .base_segmenter import BaseSegmenter


# TODO: RandomSegmenter takes a random seed.


class RandomSegmenter(BaseSegmenter):
    """A random segmenter.

    Segmentation is predicted at random at each potential word
    boundary independently for a given probability. No training is required.
    """

    def __init__(self, *, prob):
        """Initialize a random segmenter.

        Parameters
        ----------
        prob : float
            The probability from [0, 1) that segmentation occurs between
            two symbols.

        Raises
        ------
        ValueError
            If prob is outside [0, 1).
        """
        self.prob = prob
        if not 0 <= self.prob < 1:
            raise ValueError(f"prob must be from [0, 1): {prob}")

    def fit(self, sents):
        raise NotImplementedError("No training needed for RandomSegmenter")

    def _predict_sent(self, sent_str):
        segment_or_not = [
            self.prob > random.random() for _ in range(len(sent_str) - 1)  # nosec
        ]
        boundaries = [i + 1 for i, seg in enumerate(segment_or_not) if seg]
        sent = []
        for i, j in zip([0] + boundaries, boundaries + [len(sent_str)]):
            word = sent_str[i:j]
            sent.append(word)
        return sent



from abc import ABCMeta, abstractmethod


class BaseSegmenter(metaclass=ABCMeta):
    """Base class of a word segmentation model."""

    @abstractmethod
    def fit(self, sents):
        """Train the model with the input segmented sentences.

        No cleaning or preprocessing (e.g, normalizing upper/lowercase,
        tokenization) is performed on for the training data.

        For consistency across all word segmentation model classes,
        the input to training must be an iterable of segmented sentences,
        even though some models could have used simpler data structure
        for training (e.g., just a wordlist for the longest string matching
        model).

        Parameters
        ----------
        sents : Iterable[Iterable[str]]
            An iterable of segmented sentences
        """
        pass

    def predict(self, sent_strs):
        """Segment the given unsegmented sentences.

        Parameters
        ----------
        sent_strs : Iterable[str]
            An iterable of unsegmented sentences

        Returns
        -------
        Iterable[Iterable[str]]
            A generator of segmented sentences
        """
        return (self._predict_sent(s) for s in sent_strs)

    @abstractmethod
    def _predict_sent(self, sent_str):
        """Segment the given unsegmented sentence.

        Parameters
        ----------
        sent_str : str
            An unsegmented sentence

        Returns
        -------
        Iterable[str]
            A segmented sentence
        """
        pass



try:
    from importlib.metadata import version
except ModuleNotFoundError:
    # For Python 3.7
    from importlib_metadata import version

from .longest_string_matching import LongestStringMatching
from .random_segmenter import RandomSegmenter


__version__ = version("wordseg")

__all__ = ["__version__", "RandomSegmenter", "LongestStringMatching"]
